/* Malloc Debugging Library
 * The source file
 * Michael McTernan
 * mm7323@Bris.Ac.Uk
 */

/* NOTE: Don't include Dbmalloc.h in this file! */

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef _REENTRANT
 #include <pthread.h>
#endif 

#include "DbMallocOpts.h"

typedef struct MemList {
	void   *Ptr;
        size_t  Size;
	char   *File;
	int     Line;
	struct  MemList *Next;
} MemList_t;

/* Globals */
static MemList_t *MemList=NULL;
#ifdef _REENTRANT
  static pthread_mutex_t MemMutex=PTHREAD_MUTEX_INITIALIZER;
#endif

/* Prototypes */
static void RemoveEntry (void *Ptr);
static void AddEntry (void *Ptr,size_t Size, char *File, int Line);

void *DbMalloc(size_t Size, char *File, int Line)
{
	void *NewMem=malloc(Size);

	AddEntry (NewMem,Size,File, Line);

	return NewMem;
}

void *DbCalloc(size_t Size, size_t Num,char *File, int Line)
{
	void *NewMem=calloc(Size,Num);

	AddEntry(NewMem,Size*Num,File, Line);

	return NewMem;
}

char *DbStrdup(char *String, char *File, int Line)
{
	char *NewStr=strdup(String);

	AddEntry(NewStr,strlen(String)+1,File, Line);

	return NewStr;
}

void *DbRealloc(void *OldPtr, size_t NewSize, char *File, int Line)
{
	void *NewPtr=realloc(OldPtr,NewSize);

	RemoveEntry(OldPtr);
	AddEntry(NewPtr,NewSize,File, Line);

	return NewPtr;
}

void DbFree(void *Mem, char *File, int Line)
{
	
#ifdef WARN_FREENULL
	if(Mem==NULL)
		fprintf(stderr,"%s: %d  Warning, freeing NULL\n",File,Line);
#endif
	
	if(Mem==NULL)
		return;

#ifdef DEAD_BEEF
        *(char *)Mem=(char)DEAD_BEEF;
#endif

	free(Mem);

	RemoveEntry(Mem);

	return;
}

void DbListUsed(void)
{
	MemList_t *Current;
#ifdef MORE
	int Line=0;
#endif

#ifndef DB_ON
	return;
#endif

	fprintf(stderr,"-----------\n");
	fprintf(stderr,"Allocated memory:-\n");
	fprintf(stderr,"-----------\n");
	
#ifdef _REENTRANT
	pthread_mutex_lock(&MemMutex);	
#endif	
	Current=MemList;

	while(Current!=NULL) {
		fprintf(stderr,"%x %5d %s:%d\n",
                               (unsigned int)Current->Ptr,
                               Current->Size,
                               Current->File,
                               Current->Line);
		Current=Current->Next;
#ifdef MORE
		Line++;
		if(Line>MORE) {
			fprintf(stderr,"Press return key to continue...\n");
			getchar();
			Line=0;
		}
#endif
	}

#ifdef _REENTRANT
	pthread_mutex_unlock(&MemMutex);	
#endif	

	fprintf(stderr,"-----------\n");

#ifdef MORE
	fprintf(stderr,"Press return to continue...\n");
	getchar();
#endif
}

void DbFreeUsed(void)
{
	MemList_t *Current;

#ifndef DB_ON
	return;
#endif

#ifdef _REENTRANT
	pthread_mutex_lock(&MemMutex);	
#endif	
	Current=MemList;

	while(Current!=NULL) {
		free(Current->Ptr);
		free(Current->File);
		free(Current);
		MemList=Current->Next;
		free(Current);
		Current=MemList;
	}

	MemList=NULL;

#ifdef _REENTRANT
	pthread_mutex_unlock(&MemMutex);	
#endif

}

static void AddEntry (void *Ptr,size_t Size,char *File, int Line)
{
	MemList_t *Current;
	
#ifdef _REENTRANT
	pthread_mutex_lock(&MemMutex);	
#endif	

	Current=calloc(sizeof(MemList_t),1);

	if(Current==NULL) {
		fprintf(stderr,"DbMalloc: Error, the real calloc failed\n");
		abort();
	}

	Current->Ptr=Ptr;
	Current->Size=Size;
	Current->File=strdup(File);
	Current->Line=Line;
	Current->Next=MemList;

	MemList=Current;

#ifdef _REENTRANT
	pthread_mutex_unlock(&MemMutex);	
#endif	
	return;
}

static void RemoveEntry (void *Ptr)
{
	MemList_t *Current,*Prev;

#ifdef _REENTRANT
	pthread_mutex_lock(&MemMutex);	
#endif	
	Prev=NULL;
	Current=MemList;

	while(Current!=NULL) {

		if(Current->Ptr==Ptr) {
			free(Current->File);
			
			if(Prev!=NULL) 
				Prev->Next=Current->Next;
			else 
				MemList=Current->Next;
				
			free(Current);			
#ifdef _REENTRANT
			pthread_mutex_unlock(&MemMutex);	
#endif
			return;
		}
		Current=Current->Next;
	}
	
#ifdef _REENTRANT
	pthread_mutex_unlock(&MemMutex);	
#endif	
	return;

}

